/*  XMMS - ALSA output plugin 
 *  Copyright (C) 2001 Matthieu Sozeau <mattam@netcourrier.com>
 * 
 *  Contains code 
 *      Copyright (C) 1998-2001  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *      Copyright (C) 1999-2001  Hvard Kvlen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "alsa.h"

#define error(str) do { printf(str); printf("\n"); exit(-1); } while(0);

static snd_pcm_t *alsa_pcm = NULL;
static snd_pcm_t *alsa_pcm_slave = NULL;
static snd_pcm_hw_params_t *hwparams = NULL;
static snd_pcm_sw_params_t *swparams = NULL;
static snd_pcm_format_t alsa_format;
static snd_pcm_status_t *alsa_status = NULL;

static snd_output_t *logs = NULL;

static unsigned int alsa_rate;
static unsigned int alsa_channels = 2;
static int alsa_buffer_size;
static int alsa_buffer_time;
static int alsa_bits_per_sample;
static int alsa_frame_size;
static int written_frames;
static int alsa_bps = 0;
static long int alsa_total_written = 0;
static char device[7] = "hw:0,0";

//Surround
static void *datas = NULL;
static short int surround = 0;

//Set/Get volume
static long int alsa_min_vol, alsa_max_vol;
static snd_pcm_info_t *info = NULL;  
static snd_mixer_selem_id_t *pcm_elt = NULL;
static snd_mixer_elem_t *pcm_element = NULL, *pcm_element_surround = NULL;
static snd_mixer_t *mixer = NULL;
static char *card_id = "hw:0";

static int alsa_card = 0, alsa_device = 0, alsa_subdevice = 0;

//Loop
static int flush = -1, written = 0, going = 0, pause_time;
static gboolean paused = FALSE;
static int realtime;

static gint output_bytes = 0;
static gpointer buffer;
static pthread_t buffer_thread;

void alsa_setup_mixer(void);

static void alsa_setup(AFormat fmt,gint rate, gint nch);
static void *alsa_loop(void *arg);
void alsa_write(gpointer data, gint length);

int err, count, i;

gint alsa_playing(void)
{
	if(!going || paused)
		return FALSE;
  
	return(snd_pcm_state(alsa_pcm) == SND_PCM_STATE_RUNNING);
}

gint alsa_free(void)
{
	if (paused)
		return 0;
	else
		return 100000;
}

void alsa_pause(short p)
{
	if(p)
		snd_pcm_drop(alsa_pcm);

	snd_pcm_prepare(alsa_pcm);

	paused = p;
}


void alsa_close(void)
{
	if(alsa_cfg.debug) printf("Closing device\n");
	if (!going)
		return;
	going = 0;

	snd_mixer_selem_id_free(pcm_elt);
	snd_pcm_close(alsa_pcm);
	alsa_pcm = NULL;
	if(alsa_cfg.surround) free(datas);      
}

void alsa_flush(gint time)
{
	alsa_total_written = (unsigned long int) time * alsa_bps;
}

static long int a, b;
void alsa_setup_mixer(void)
{ 
	snd_mixer_open(&mixer, 0);
	snd_mixer_attach(mixer, card_id);
	snd_mixer_selem_register(mixer, NULL, NULL);
	snd_mixer_load (mixer);
	snd_mixer_selem_id_malloc(&pcm_elt);

	if(alsa_cfg.use_master) 
	{
		pcm_element = snd_mixer_first_elem(mixer);  
		snd_mixer_selem_get_id(pcm_element, pcm_elt);
		while(strcmp(snd_mixer_selem_id_get_name(pcm_elt), "Master"))
		{
			pcm_element = snd_mixer_elem_next(pcm_element);
			snd_mixer_selem_get_id(pcm_element, pcm_elt);
		}
	}
	else
	{
		pcm_element = snd_mixer_first_elem(mixer);  
		snd_mixer_selem_get_id(pcm_element, pcm_elt);
		while(strcmp(snd_mixer_selem_id_get_name(pcm_elt), "PCM"))
		{
			pcm_element = snd_mixer_elem_next(pcm_element);
			snd_mixer_selem_get_id(pcm_element, pcm_elt);
		}
	}
	
	snd_mixer_selem_get_playback_volume_range(pcm_element, &alsa_min_vol, &alsa_max_vol);
	snd_mixer_selem_set_playback_volume_range(pcm_element, 0, 100);
  
	snd_mixer_selem_get_playback_volume(pcm_element, SND_MIXER_SCHN_FRONT_LEFT, &a);
	snd_mixer_selem_get_playback_volume(pcm_element, SND_MIXER_SCHN_FRONT_RIGHT, &b);
	alsa_set_volume(a * 100 / alsa_max_vol, b * 100 / alsa_max_vol);
}

static long int volleft, volright;
void alsa_get_volume(int *l, int *r)
{
	if(!pcm_element)
	{
		*l = *r = 0;
		return;
	}
	if(snd_mixer_selem_has_playback_volume(pcm_element))
	{
		snd_mixer_selem_get_playback_volume(pcm_element, SND_MIXER_SCHN_FRONT_LEFT, &volleft);
		snd_mixer_selem_get_playback_volume(pcm_element, SND_MIXER_SCHN_FRONT_RIGHT, &volright);
		*l = volleft;
		*r = volright; 
	}
	else 
	{
		error("alsa_get_volume: Wrong PCM channel!");
	}
}



void alsa_set_volume(int l, int r)
{
	if(!pcm_element) return;

	if(snd_mixer_selem_has_playback_volume(pcm_element))
	{
		snd_mixer_selem_set_playback_volume(pcm_element, SND_MIXER_SCHN_FRONT_LEFT, l);
		snd_mixer_selem_set_playback_volume(pcm_element, SND_MIXER_SCHN_FRONT_RIGHT, r);
	}
	else error("alsa_set_volume: Wrong PCM channel!");
}


gint alsa_get_output_time(void)
{
	if(going) return(alsa_total_written / alsa_bps); 
	return(0);
}

gint alsa_get_written_time(void)
{
	return(alsa_total_written / alsa_bps);
}


void alsa_write(gpointer data, gint length)
{
	if(paused) return;
	  
	if(surround)
	{
		for(i = 0; i < length; i += 2)
		{
			memcpy(datas + 2*i, data + i, 2);
			memcpy(datas + 2*i + 2, data + i, 2);
		}
		data = datas;
		length = length*2;
	}
	while(length > 0)
	{          
		written_frames = snd_pcm_writei(alsa_pcm, data, length / alsa_frame_size);
		// printf("written frames: %i, %s\n", written_frames, snd_strerror(written_frames));
		// printf("Wr: %i, prop: %i, framesize: %i\n", written_frames, length/alsa_frame_size, alsa_frame_size);
		if(written_frames > 0)
		{
			written = written_frames * alsa_frame_size;
			alsa_total_written += written;
			length -= written;
			data += written;
		}
		else if (written_frames == -EAGAIN || (written_frames > 0 && written_frames < (length / alsa_frame_size)))
		{
			snd_pcm_wait(alsa_pcm, 1000);
		}
		else if(written_frames == -EPIPE)
		{
			if(alsa_cfg.debug) printf("xrun error");	
			snd_pcm_status_alloca(&alsa_status);
			if (snd_pcm_status(alsa_pcm, alsa_status)<0) {
				error("status error");
			}
			if (snd_pcm_status_get_state(alsa_status) == SND_PCM_STATE_XRUN) {
				if (alsa_cfg.debug) {
					printf("Status:\n");
					snd_pcm_status_dump(alsa_status, logs);
				}
				if ((snd_pcm_prepare(alsa_pcm))<0) {
					error("xrun: prepare error");
				}
			}
			else
			{
				error("read/write error");
			}
		}
	}
}

gint alsa_open(AFormat fmt, gint rate, gint nch)
{
	if(alsa_cfg.debug) printf("Opening device\n");

	if(alsa_pcm) alsa_close();
	if(alsa_cfg.debug) snd_output_stdio_attach(&logs, stdout, 0);
  
	if(alsa_cfg.surround) surround = 1;
	else surround = 0;

	alsa_setup(fmt,rate,nch);
	alsa_setup_mixer();

	flush = -1;
	written = output_bytes = 0;
	alsa_total_written = 0;
	going = 1;
	snd_pcm_prepare(alsa_pcm);  
	paused = FALSE;


	/*  if(!(realtime = xmms_check_realtime_priority()))
	    {
	    printf("Realtime? %i\n", realtime);
	    printf("Buffer size: %i\n", alsa_buffer_size);  
	    pthread_create(&buffer_thread, NULL, alsa_loop, NULL);
	    }*/
    
	return 1;
}

static void swap_endian(guint16 *data, int length)
{
	int i;
	for (i = 0; i < length; i += 2, data++)
		*data = GUINT16_SWAP_LE_BE(*data);
}



#define NOT_NATIVE_ENDIAN ((IS_BIG_ENDIAN &&				\
			   (alsa_format == SND_PCM_FORMAT_S16_LE ||		\
			    alsa_format == SND_PCM_FORMAT_U16_LE)) ||	\
			  (!IS_BIG_ENDIAN &&				\
			   (alsa_format == SND_PCM_FORMAT_S16_BE ||		\
			    alsa_format == SND_PCM_FORMAT_U16_BE)))


#define RESAMPLE_STEREO(sample_type)				\
do {								\
	const gint shift = sizeof (sample_type);		\
        gint i, in_samples, out_samples, x, delta;		\
	sample_type *inptr = (sample_type *)ob, *outptr;	\
	guint nlen = (((length >> shift) * espeed) / speed);	\
	if (nlen == 0)						\
		break;						\
	nlen <<= shift;						\
	if (NOT_NATIVE_ENDIAN)					\
		swap_endian(ob, length);			\
	if(nlen > nbuffer_size)					\
	{							\
		nbuffer = g_realloc(nbuffer, nlen);		\
		nbuffer_size = nlen;				\
	}							\
	outptr = (sample_type *)nbuffer;			\
	in_samples = length >> shift;				\
        out_samples = nlen >> shift;				\
	delta = (in_samples << 12) / out_samples;		\
	for (x = 0, i = 0; i < out_samples; i++)		\
	{							\
		gint x1, frac;					\
		x1 = (x >> 12) << 12;				\
		frac = x - x1;					\
		*outptr++ =					\
			(sample_type)				\
			((inptr[(x1 >> 12) << 1] *		\
			  ((1<<12) - frac) +			\
			  inptr[((x1 >> 12) + 1) << 1] *	\
			  frac) >> 12);				\
		*outptr++ =					\
			(sample_type)				\
			((inptr[((x1 >> 12) << 1) + 1] *	\
			  ((1<<12) - frac) +			\
			  inptr[(((x1 >> 12) + 1) << 1) + 1] *	\
			  frac) >> 12);				\
		x += delta;					\
	}							\
	if (NOT_NATIVE_ENDIAN)					\
		swap_endian(nbuffer, nlen);			\
	alsa_write(nbuffer, nlen);			        \
} while (0)


#define RESAMPLE_MONO(sample_type)				\
do {								\
	const gint shift = sizeof (sample_type) - 1;		\
        gint i, x, delta, in_samples, out_samples;		\
	sample_type *inptr = (sample_type *)ob, *outptr;	\
	guint nlen = (((length >> shift) * espeed) / speed);	\
	if (nlen == 0)						\
		break;						\
	nlen <<= shift;						\
	if (NOT_NATIVE_ENDIAN)					\
		swap_endian(ob, length);			\
	if(nlen > nbuffer_size)					\
	{							\
		nbuffer = g_realloc(nbuffer, nlen);		\
		nbuffer_size = nlen;				\
	}							\
	outptr = (sample_type *)nbuffer;			\
	in_samples = length >> shift;				\
        out_samples = nlen >> shift;				\
	delta = ((length >> shift) << 12) / out_samples;	\
	for (x = 0, i = 0; i < out_samples; i++)		\
	{							\
		gint x1, frac;					\
		x1 = (x >> 12) << 12;				\
		frac = x - x1;					\
		*outptr++ =					\
			(sample_type)				\
			((inptr[x1 >> 12] * ((1<<12) - frac) +	\
			  inptr[(x1 >> 12) + 1] * frac) >> 12);	\
		x += delta;					\
	}							\
	if (NOT_NATIVE_ENDIAN)					\
		swap_endian(nbuffer, nlen);			\
	alsa_write(nbuffer, nlen);			        \
} while (0)


static gint alsa_downsample(gpointer ob, guint length, guint speed, guint espeed)
{
	guint w = 0;
	static gpointer nbuffer = NULL;
	static gint nbuffer_size = 0;

	switch (alsa_format)
	{
	case SND_PCM_FORMAT_S16_BE:
	case SND_PCM_FORMAT_S16_LE:
		if (alsa_channels == 2)
			RESAMPLE_STEREO(gint16);
		else
			RESAMPLE_MONO(gint16);
		break;
	case SND_PCM_FORMAT_U16_BE:
	case SND_PCM_FORMAT_U16_LE:
		if (alsa_channels == 2)
			RESAMPLE_STEREO(guint16);
		else
			RESAMPLE_MONO(guint16);
		break;
	case SND_PCM_FORMAT_S8:
		if (alsa_channels == 2)
			RESAMPLE_STEREO(gint8);
		else
			RESAMPLE_MONO(gint8);
		break;
	case SND_PCM_FORMAT_U8:
		if (alsa_channels == 2)
			RESAMPLE_STEREO(guint8);
		else
			RESAMPLE_MONO(guint8);
		break;
	}
	return w;
}

static void alsa_setup(AFormat fmt,gint rate, gint channels)
{
	int chosen_rate;

	switch(fmt)
	{
	case FMT_U8:
		alsa_format = SND_PCM_FORMAT_U8;
		break;
	case FMT_S8:
		alsa_format = SND_PCM_FORMAT_S8;
		break;
	case FMT_U16_LE:
		alsa_format = SND_PCM_FORMAT_U16_LE;
		break;
	case FMT_U16_BE:
		alsa_format = SND_PCM_FORMAT_U16_BE;
		break;
	case FMT_U16_NE:
#ifdef AFMT_U16_NE
		error("Unrecognized format: AFMT_U16_NE");
		break;
#else
#ifdef WORDS_BIGENDIAN
		alsa_format = SND_PCM_FORMAT_U16_BE;
#else
		alsa_format = SND_PCM_FORMAT_U16_LE;
#endif
#endif
		break;
	case FMT_S16_LE:
		alsa_format = SND_PCM_FORMAT_S16_LE;
		break;
	case FMT_S16_BE:
		alsa_format = SND_PCM_FORMAT_S16_BE;
		break;
	case FMT_S16_NE:
#ifdef AFMT_S16_NE
		error("Unrecognized format: AFMT_S16_NE");
		break;
#else
#ifdef WORDS_BIGENDIAN
		alsa_format = SND_PCM_FORMAT_S16_BE;
#else
		alsa_format = SND_PCM_FORMAT_S16_LE;
#endif
#endif
		break;
	}

	alsa_rate = rate;
	alsa_channels = channels;
  
	if(alsa_cfg.use_user_device)
	{
		if(alsa_cfg.debug) printf("Opening user device: %s\n", alsa_cfg.user_device);
		snd_pcm_open(&alsa_pcm, alsa_cfg.user_device, SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK);      
	}
	else if(surround)
	{
		alsa_channels = 4;
		datas = (int *) malloc(10000);
		if(alsa_cfg.debug) printf("Surround!\n");
		snd_pcm_open(&alsa_pcm,"surround40", SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK);            
	}
	else
	{
		device[3] = (char) (alsa_cfg.audio_card + 48);
		device[5] = (char) (alsa_cfg.audio_device + 48);
		if(alsa_cfg.debug) printf("Opening device: %s\n", device);
		snd_pcm_open(&alsa_pcm, device, SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK);
	}

	if(alsa_cfg.debug) 
	{
		if(!info) snd_pcm_info_malloc(&info);
		snd_pcm_info(alsa_pcm, info);
		alsa_card =  snd_pcm_info_get_card(info);
		alsa_device = snd_pcm_info_get_device(info);
		alsa_subdevice = snd_pcm_info_get_subdevice(info);
		printf("Card %i, Device %i, Subdevice %i\n",  alsa_card,   alsa_device,  alsa_subdevice);  
		printf("setup_audio_params\n");
	}

	snd_pcm_hw_params_malloc(&hwparams);
	if(!hwparams) error("No hw params allocated");
  
	snd_pcm_hw_params_any(alsa_pcm, hwparams);

	snd_pcm_hw_params_set_access(alsa_pcm, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED);
	snd_pcm_hw_params_set_format(alsa_pcm, hwparams, alsa_format);
	snd_pcm_hw_params_set_channels(alsa_pcm, hwparams, alsa_channels);
	chosen_rate = snd_pcm_hw_params_set_rate_near(alsa_pcm, hwparams, alsa_rate, 0);
	if(chosen_rate < 0)
		error("Cannot set rate.");
	if( chosen_rate!=alsa_rate ) {
		if(alsa_cfg.debug) {
			printf("Rate doesn't match (requested %iHz, get %iHz)\n",
			       alsa_rate, chosen_rate);
			printf("Workaround, creating plugin...\n");
		}
		alsa_pcm_slave = alsa_pcm;
		snd_pcm_rate_open(&alsa_pcm,
				  "",
				  alsa_format,
				  chosen_rate,
				  alsa_pcm_slave,
				  1);
		snd_pcm_hw_params_any(alsa_pcm, hwparams);

		snd_pcm_hw_params_set_access(alsa_pcm, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED);
		snd_pcm_hw_params_set_format(alsa_pcm, hwparams, alsa_format);
		snd_pcm_hw_params_set_channels(alsa_pcm, hwparams, alsa_channels);
		chosen_rate =
			snd_pcm_hw_params_set_rate_near(alsa_pcm, hwparams, alsa_rate, 0);
		if(chosen_rate!=alsa_rate)
			error("Cannot set rate.");
	}
  

	alsa_buffer_time = snd_pcm_hw_params_set_buffer_time_near(alsa_pcm, hwparams,  alsa_cfg.buffer_time * 100, 0);
	if (alsa_buffer_time < 0) error("Buffer time < 0");
  
	if (snd_pcm_hw_params_set_period_time_near(alsa_pcm, hwparams, alsa_buffer_time / 4, 0) < 0) error("Period time < 0");

	snd_pcm_hw_params(alsa_pcm, hwparams);
	/*  
	    Don't know why but segfaults in xmms when returning (even Succes)
	    if ( < 0) {
	    if(alsa_cfg.debug) snd_pcm_hw_params_dump(hwparams, logs);
	    error("Unable to install hw params");
	    }*/

	alsa_buffer_size = snd_pcm_hw_params_get_buffer_size(hwparams);
	if (snd_pcm_hw_params_get_period_size(hwparams, 0) == alsa_buffer_size) exit(-1);

	swparams = (snd_pcm_sw_params_t *) malloc(snd_pcm_sw_params_sizeof());

	snd_pcm_sw_params_current(alsa_pcm, swparams);
	if(!swparams) error("No sw params allocated");

	if (snd_pcm_sw_params(alsa_pcm, swparams) < 0)
	{
		if(alsa_cfg.debug) snd_pcm_sw_params_dump(swparams, logs);
		error("Unable to install sw params");
	}

	if(alsa_cfg.debug)  snd_pcm_dump(alsa_pcm, logs);
  
	alsa_bits_per_sample = snd_pcm_format_physical_width(alsa_format); 
	alsa_frame_size = (alsa_bits_per_sample * alsa_channels) / 8;
	alsa_bps = (alsa_rate * alsa_frame_size) / 1000;

	if(alsa_cfg.debug)
	{
		printf("Device setup: buffer time: %i, size: %i. bits per sample: %i, frame size: %i\n", alsa_buffer_time, alsa_buffer_size, alsa_bits_per_sample, alsa_frame_size);
		printf("Bps: %i\n", alsa_bps);
	}
}
